"""
URL configuration for galactic_bakeshop project.
"""

from __future__ import annotations

from django.contrib import admin
from django.urls import include, path

urlpatterns = [
    path("admin/", admin.site.urls),
    path("", include("treats.urls")),
]
